#include "cnetheader.h"
#include "statistics.h"

/*  The cnet network simulator (v1.7.7)
    Copyright (C) 1992-2002, Chris McDonald

    Chris McDonald, chris@cs.uwa.edu.au
    Department of Computer Science & Software Engineering
    The University of Western Australia,
    Crawley, Western Australia, 6009
    PH: +61 8 9380 2533, FAX: +61 8 9380 1089.

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#define	POLL_FREQ		10			/* in ms */
#define	PINGX_FREQ		20			/* in ms */

#define	SCHEDULER_RAND48_SEED	2675255


#define	CONTEXT_SWITCH()	if(THISNODE != SWAPPED_IN) context_switch()

#define	CHECK_EVENTS_BREAK()	if(events_left == 0) break
#define	CHECK_EVENTS_RETURN()	if(events_left == 0) return(events_left)
#define	CHECK_STATE_RETURN()	if(cnet_state != STATE_RUNNING) \
				    return(events_left)

static	NODE		*np;			/* always &NP[THISNODE] */

#if	defined(USE_WIN32)
struct timezone {
    int  tz_minuteswest;			/* minutes W of Greenwich */
    int  tz_dsttime;				/* type of dst correction */
};

int gettimeofday(struct timeval *tv, struct timezone *tz);
#endif

static	long			STARTSEC		= 0L;
static	long			running_msec		= 0L;
static	int			events_handled, events_left;
static	int			cntx=0, SWAPPED_IN;
static	int			report_freq		= 0;
static	int			tell_nnodes		= 0;

static	unsigned short		xsubi[3]		= {0,0,0};

#if	defined(USE_TCLTK)

extern	void draw_link(int, int, int);
extern	void draw_map(int, int);

#define	DRAW_NODE_ICON()	if(Wflag) draw_node_icon(FALSE,THISNODE)
#define	DRAW_LINK(l)		if(Wflag) { \
				    draw_link(l, TRUE, CANVAS_FATLINK/2); \
				    draw_node_icon(FALSE,LP[l].endmin); \
				    draw_node_icon(FALSE,LP[l].endmax); }
#define	DRAW_MAP()		if(Wflag) draw_map(0,0)

#if	defined(USE_TCLTK)
#define	PING_TCLTK()		if(Wflag) tcltk_notify_dispatch()
#endif

#else	/* !defined(USE_TCLTK) */

#define	PING_TCLTK()
#define	DRAW_NODE_ICON()
#define	DRAW_LINK(l)
#define	DRAW_MAP()

#endif	/* defined(USE_TCLTK) */


/* ------------------------------------------------------------------------ */


typedef enum {
    SE_NULL = 0,
    SE_APPLICATIONREADY,	SE_TIMER,
    SE_NODESTATE,		SE_LINKSTATE,
    SE_REPORT,			SE_TIMESUP,
#if	defined(USE_TCLTK)
    SE_DEBUGBUTTON,		SE_KEYBOARDREADY,
    SE_DRAWFRAME,		SE_PINGX
#endif
} SCHEDULE;

typedef struct _sq {
    struct _sq		*next;
    SCHEDULE		se;
    int			node;
    CnetEvent		ne;
    long		ms;
    CnetTimer		timer;
    CnetData		data;
} SCHEDULE_Q;


static	SCHEDULE_Q	GSQ	= { (SCHEDULE_Q *)NULL };
static	int		nqueued	= 0;


/* --------------------------------------------------------------------- */


/*  We keep the Global Schedule Queue sorted by:

	time	- so that events occur in order,
	node	- to minimize context switches,
	ne	- so that ack frames arrive before their timeouts
 */

static void insert_event(SCHEDULE se, CnetEvent	ne, int	node, long abs_ms,
			 CnetTimer timer, CnetData data)
{
    SCHEDULE_Q	*s	= &GSQ;
    SCHEDULE_Q	*t;

    t		= NEW(SCHEDULE_Q);
    t->se	= se;
    t->node	= node;
    t->ne	= ne;
    t->ms	= abs_ms;
    t->timer	= timer;
    t->data	= data;

    while(s->next) {
	if(abs_ms  < s->next->ms)
	    break;
	if(abs_ms == s->next->ms) {
	    if(node  < s->next->node)
		break;
	    if(node == s->next->node && (int)ne < (int)s->next->ne)
		break;
	}
	s	= s->next;
    }
    t->next	= s->next;
    s->next	= t;
    ++nqueued;
}

void schedule_event(CnetEvent ne, int node, long rel_ms,
		    CnetTimer timer, CnetData data)
{
    SCHEDULE	se;

    if(node < 0) {		/* this event is for a link, not a node */
	se	= SE_LINKSTATE;
	node	= -node-1;
    }
    else if(IS_TIMER(ne))
	se	= SE_TIMER;
#if	defined(USE_TCLTK)
    else if(ne == EV_DRAWFRAME)
	se	= SE_DRAWFRAME;
    else if(IS_DEBUG(ne))
	se	= SE_DEBUGBUTTON;
    else if(ne == EV_KEYBOARDREADY)
	se	= SE_KEYBOARDREADY;
#endif
    else if(ne == EV_REBOOT)
	se	= SE_NODESTATE;
    else
	return;

    insert_event(se,ne, node, TIMENOW_in_MSEC + (rel_ms==0?1:rel_ms),
		    timer, data);
}

int unschedule_timer(CnetTimer timer, CnetData *datawas)
{					/* reqd by: CNET_stop_timer() */
    SCHEDULE_Q	*t;
    SCHEDULE_Q	*s	= &GSQ;

    while((t = s->next)) {
	if(timer == t->timer && THISNODE == t->node && IS_TIMER(t->ne)) {
	    if(datawas) {
		*datawas	= t->data;
		return(TRUE);
	    }
	    s->next	= t->next;
	    free((char *)t);
	    --nqueued;
	    return(TRUE);
	}
	s	= s->next;
    }
    return(FALSE);
}

void filter_node_events(int n)	/* reqd by: node_menu.c */
{
    SCHEDULE_Q	*s	= &GSQ;
    SCHEDULE_Q	*t;

    while((t = s->next)) {
	if(t->node == n && t->se != SE_LINKSTATE) {
	    s->next	= t->next;
	    free((char *)t);
	    --nqueued;
	}
	s	= s->next;
    }
}



/* ---------- Calculate times of node and link failure and repair --------- */


extern long poisson_msecs(long, unsigned short *);

static long node_mtbf()
{
    long t1	= WHICH(np->nattr.nodemtbf, DEFAULTNODE.nodemtbf);
    long t2	= WHICH(np->nattr.nodemtbf, DEFAULTNODE.nodemtbf);
    long extent;

         if(t1 == 0)
	extent = t2;
    else if(t2 == 0 || t1 < t2)
	extent = t1;
    else
	extent = t2;

    return (extent == 0 ? 0 : poisson_msecs(extent, xsubi));
}

static long node_mttr()
{
    long t1	= WHICH(np->nattr.nodemttr, DEFAULTNODE.nodemttr);
    long t2	= WHICH(np->nattr.nodemttr, DEFAULTNODE.nodemttr);

    return (poisson_msecs(t1 > t2 ? t1 : t2, xsubi));
}


static long link_mtbf(int l)
{
    long t1	= WHICH(LP[l].lattrmin.linkmtbf, DEFAULTLINK.linkmtbf);
    long t2	= WHICH(LP[l].lattrmax.linkmtbf, DEFAULTLINK.linkmtbf);
    long extent;

       if(t1 == 0)
	extent = t2;
    else if(t2 == 0 || t1 < t2)
	extent = t1;
    else
	extent = t2;

    return (extent == 0 ? 0 : poisson_msecs(extent, xsubi));
}

static long link_mttr(int l)
{
    long t1 = WHICH(LP[l].lattrmin.linkmttr, DEFAULTLINK.linkmttr);
    long t2 = WHICH(LP[l].lattrmax.linkmttr, DEFAULTLINK.linkmttr);

    return (poisson_msecs(t1 > t2 ? t1 : t2, xsubi));
}


/* ---------------------- sundry scheduling functions ----------------- */


int calc_time_delta(int newsec, int newmsec)	   /* for timers.c */
{
    return (newsec * 1000 + newmsec) - (STARTSEC * 1000  + TIMENOW_in_MSEC);
}

/* ------------------------------------------------------------------------ */


static void swapout_data(int n)
{
    NODE *np	= &NP[n];
    int	 i;

    for(i=0 ; i<NDATASEGS ; ++i)
	if(np->length_data[i])
	    memcpy(np->private_data[i], np->incore_data[i], np->length_data[i]);
}

static void swapin_data(int n)
{
    NODE *np	= &NP[n];
    int	 i;

    for(i=0 ; i<NDATASEGS ; ++i)
	if(np->length_data[i])
	    memcpy(np->incore_data[i], np->private_data[i], np->length_data[i]);
}

static void reboot_data(int n)
{
    NODE *np	= &NP[n];
    int  i;

    for(i=0 ; i<NDATASEGS ; ++i)
	if(np->length_data[i])
	    memcpy(np->incore_data[i],np->original_data[i],np->length_data[i]);
}

/* ------------------------------------------------------------------------ */

static void	context_switch()
{
    CnetLinkinfo	*li;
    LINKATTR		*thislinkattr;
    int			l, whichlink;

/* swap out SWAPPED_IN process, swap in THISNODE process */

    NP[SWAPPED_IN].os_errno	= errno;
    NP[SWAPPED_IN].cnet_errno	= cnet_errno;

    swapout_data(SWAPPED_IN);
    swapin_data(THISNODE);

    nodeinfo.nodenumber	= THISNODE;
    strcpy(nodeinfo.nodename, np->nodename);
    nodeinfo.nodetype	= np->nodetype;
    nodeinfo.address	= np->nattr.address;
    nodeinfo.nlinks	= np->nlinks;

    nodeinfo.minmessagesize	=
	    WHICH(np->nattr.minmessagesize, DEFAULTNODE.minmessagesize);
    nodeinfo.maxmessagesize	=
	    WHICH(np->nattr.maxmessagesize, DEFAULTNODE.maxmessagesize);
    nodeinfo.messagerate	=
	    WHICH(np->nattr.messagerate, DEFAULTNODE.messagerate);

    for(l=0 ; l<np->nlinks ; ++l) {
	whichlink	= np->links[l];
	if( LP[whichlink].endmin == THISNODE)
	    thislinkattr	= &(LP[whichlink].lattrmin);
	else
	    thislinkattr	= &(LP[whichlink].lattrmax);

	li			= &linkinfo[l+1];
	li->linkup		= LP[whichlink].linkup;
	li->bandwidth		=
	    WHICH(thislinkattr->bandwidth, DEFAULTLINK.bandwidth);
	li->costperbyte		=
	    WHICH(thislinkattr->costperbyte, DEFAULTLINK.costperbyte);
	li->costperframe	=
	    WHICH(thislinkattr->costperframe, DEFAULTLINK.costperframe);
	li->propagationdelay	=
	    WHICH(thislinkattr->propagationdelay, DEFAULTLINK.propagationdelay);
	li->transmitbufsize	=
	    WHICH(thislinkattr->transmitbufsize, DEFAULTLINK.transmitbufsize);
    }

    errno		= np->os_errno;
    cnet_errno		= np->cnet_errno;
    NNODES		= tell_nnodes;
    SWAPPED_IN		= THISNODE;
    cntx++;
}

/* ------------------------------------------------------------------------ */

void update_clock(long inc)
{
    int	msec;

    TIMENOW_in_MSEC		+= inc;
    msec			 = TIMENOW_in_MSEC + np->clock_skew;
    nodeinfo.time_of_day.sec	 = STARTSEC + msec / 1000;
    nodeinfo.time_of_day.msec	 = msec % 1000;
    nodeinfo.time_in_ms		 = TIMENOW_in_MSEC;
}

static void HANDLER(CnetEvent ev, CnetTimer timer, CnetData data)
{
    extern void		clear_trace_names(void);

    update_clock(0L);
    HANDLING = ev;
    EV_COUNT[(int)HANDLING]++;
    inc_stats(STATS_EVENTS,0,0);

    gattr.stdio_quiet	= DEFAULTNODE.stdio_quiet	|
			  np->nattr.stdio_quiet;

    gattr.trace_events	= DEFAULTNODE.trace_all		|
			  np->nattr.trace_all		|
			  (np->nattr.trace_mask & (1<<(int)HANDLING));

    if(gattr.trace_events) {
	if(timer == NULLTIMER)
	    TRACE("enter %s(%s,NULLTIMER,data=%ld) @%ldms\n",
			np->nodename, cnet_evname[(int)HANDLING],
			data, TIMENOW_in_MSEC);
	else
	    TRACE("enter %s(%s,t=%ld,data=%ld) @%ldms\n",
			np->nodename, cnet_evname[(int)HANDLING], timer,
			data, TIMENOW_in_MSEC);
    }
    np->handler[(int)HANDLING](HANDLING, timer, data);
    clear_trace_names();
    if(gattr.trace_events)
	TRACE("leave %s(%s)\n\n", np->nodename, cnet_evname[(int)HANDLING]);

    ++events_handled;
    --events_left;
}

static void invoke_reboot()
{
    if(np->handler[(int)EV_REBOOT]) {

	extern int	reboot_application_layer(int *);
	extern int	reboot_physical_layer(void);
	extern int	reboot_stdio_layer(void);
	extern void	reboot_timer_layer(void);

	int		do_next;

	CONTEXT_SWITCH();		/* only switch out running node */
	reboot_data(THISNODE);
	filter_node_events(THISNODE);

	if(np->nodetype == NT_HOST) {
	    do_next = TIMENOW_in_MSEC + POLL_FREQ;
	    reboot_application_layer(&do_next);
	    insert_event(SE_APPLICATIONREADY, EV_APPLICATIONREADY,
				    THISNODE, do_next, NULLTIMER, 0);
	}

	reboot_physical_layer();
	reboot_stdio_layer();
	reboot_timer_layer();

	np->os_errno	= 0;
	np->cnet_errno	= ER_OK;
	cnet_errno	= ER_OK;
	HANDLER(EV_REBOOT, NULLTIMER, np->data[(int)EV_REBOOT]);
	np->runstate	= STATE_RUNNING;

/*  NOW INSERT AN EVENT IF THIS NODE IS DOOMED TO FAIL */

	do_next	= node_mtbf();
	if(do_next != 0)
	    insert_event(SE_NODESTATE, EV_REBOOT, THISNODE,
			    TIMENOW_in_MSEC + do_next, NULLTIMER, 0);
    }
    else {
	fprintf(stderr,"%s: %s rebooted without an EV_REBOOT handler!\n",
				argv0, np->nodename);
	cleanup(2);
    }
}

void invoke_shutdown(int which)
{
    np	= &(NP[THISNODE = which]);

    if(np->handler[(int)EV_SHUTDOWN]) {
	CONTEXT_SWITCH();
	HANDLER(EV_SHUTDOWN, NULLTIMER, np->data[(int)EV_SHUTDOWN]);
    }
}


/* ------------------------------------------------------------------------ */

static long running_MSEC()
{
    struct timeval	NOW;

    gettimeofday(&NOW, (struct timezone *)NULL);
    return((NOW.tv_sec-STARTSEC)*1000 + (NOW.tv_usec/1000));
}

#if	!defined(USE_WIN32)
static void Sleep(int msec)
{
    struct timeval timeout;

    timeout.tv_sec	= msec / 1000;
    timeout.tv_usec	= (msec % 1000) * 1000;
    select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &timeout);
}
#endif


/* ------------------------------------------------------------------------ */


void init_scheduler(int fflag, int Mflag, int Nflag, int Sflag)
{
/*  INIT SCHEDULER BY TAKING LOCAL COPIES OF COMMAND-LINE OPTIONS */
    report_freq		= fflag * 1000;
    running_msec	= Mflag * 60000;
    tell_nnodes		= Nflag ? _NNODES : 0;

    Sflag += SCHEDULER_RAND48_SEED;
    memcpy((char *)xsubi, (char *)&Sflag, sizeof(Sflag));
}

static void start_scheduler()
{
    int			l, n, do_next;
    struct timeval	NOW;

    gettimeofday(&NOW, (struct timezone *)NULL);
    STARTSEC		= NOW.tv_sec;
    TIMENOW_in_MSEC	= 0L;

    THISNODE		= -1;
    SWAPPED_IN		= _NNODES-1;

/*  IT IS ESSENTIAL THAT NODE 0 REBOOTS FIRST (cf. context_switch() ) */
    for(n=0 ; n<_NNODES ; ++n) {
	NP[n].runstate = STATE_REBOOTING;
	insert_event(SE_NODESTATE, EV_REBOOT, n, 0L, NULLTIMER, 0);
    }

/*  NEXT, INSERT ALL POSSIBLE LINK FAILURES */
    for(l=0 ; l<_NLINKS ; ++l) {
	do_next	= link_mtbf(l);
	if(do_next != 0)
	    insert_event(SE_LINKSTATE,EV_SHUTDOWN,l, do_next, NULLTIMER, 0);
    }

/*  FINALLY, INSERT ALL SUNDRY SCHEDULABLE EVENTS */
    insert_event(SE_TIMESUP,EV_NULL,ALLNODES,(running_msec + 1),NULLTIMER,0);
    if(report_freq > 0)
	insert_event(SE_REPORT,EV_NULL,ALLNODES,report_freq,NULLTIMER,0);
}


int schedule(int Eflag, int Tflag)
{
    extern int		poll_physical(long *when, int *node);

    SCHEDULE_Q		*sq;
    SCHEDULE		se;
    CnetEvent		ne	= EV_NULL;
    CnetTimer		timer	= NULLTIMER;
    CnetData		data;
    int			node;
    long		do_next;

    if(STARTSEC == 0L)
	start_scheduler();
    events_handled	= 0;
    cntx		= 0;
    events_left		= Eflag;
    cnet_state		= STATE_RUNNING;

#if	defined(USE_TCLTK)
    if(Wflag)
	insert_event(SE_PINGX, EV_NULL, ALLNODES,
			    TIMENOW_in_MSEC + PINGX_FREQ, NULLTIMER, 0);
#endif

    while(GSQ.next) {
	CHECK_EVENTS_RETURN();

/*  FIRSTLY, CHECK TO SEE IF WE HAVE ANY FRAMES FOR IMMEDIATE DELIVERY.
    THIS IS A BIT MORE TRICKY/UGLY THAN I'D LIKE BECAUSE WE ARE MERGING
    THE PHYSICAL FRAME QUEUE WITH THE MAIN SCHEDULING QUEUE */

	if(Tflag)
	    while(TIMENOW_in_MSEC < GSQ.next->ms) {
		Sleep(POLL_FREQ);
		TIMENOW_in_MSEC	= running_MSEC();
	    }
	TIMENOW_in_MSEC = GSQ.next->ms;
 
	if(poll_physical(&do_next, &node)) {
	    TIMENOW_in_MSEC	= do_next;
	    THISNODE		= node;
	    np			= &(NP[THISNODE]);

	    if(np->handler[ (int)EV_PHYSICALREADY ]) {
		CONTEXT_SWITCH();
		HANDLER(EV_PHYSICALREADY, NULLTIMER, (CnetData)0);
	    }
	    continue;		/* back to  while(GSQ.next)  */
	}

/*  NO PHYSICAL LAYER FRAME AVAILABLE, JUST DELIVER A NORMAL EVENT */
	
	sq		= GSQ.next;
	se		= sq->se;
	ne		= sq->ne;
	timer		= sq->timer;
	data		= sq->data;
	THISNODE	= sq->node;
	GSQ.next	= sq->next;
	free((char *)sq);
	--nqueued;

	switch ((int)se) {

	case SE_APPLICATIONREADY : {
	    extern int	poll_application(long *);

	    np		= &(NP[THISNODE]);
	    if(np->runstate == STATE_PAUSED)
		do_next = np->will_resume+1;
	    else {
		do_next = TIMENOW_in_MSEC + POLL_FREQ;
		if(poll_application(&do_next)			&&
		   np->handler[ (int)EV_APPLICATIONREADY ]) {
		    CONTEXT_SWITCH();
		    HANDLER(EV_APPLICATIONREADY, NULLTIMER, data);
		}
	    }
	    insert_event(SE_APPLICATIONREADY, EV_APPLICATIONREADY,
			    THISNODE, do_next, NULLTIMER, 0);
	    break;
	}

#if	defined(USE_TCLTK)
	case SE_DEBUGBUTTON :
	case SE_KEYBOARDREADY :
#endif
	case SE_TIMER :
	    np		= &(NP[THISNODE]);
	    if(np->runstate == STATE_PAUSED)
		insert_event(se,ne,THISNODE, np->will_resume+1,NULLTIMER,0);
	    else if(np->handler[ (int)ne ]) {
#if	defined(USE_TCLTK)
		if(se == SE_DEBUGBUTTON) {
		    int	sq1			= np->nattr.stdio_quiet;
		    int	sq2			= DEFAULTNODE.stdio_quiet;

		    np->nattr.stdio_quiet	= FALSE;
		    DEFAULTNODE.stdio_quiet	= FALSE;
			CONTEXT_SWITCH();
			HANDLER(ne, timer, data);
		    np->nattr.stdio_quiet	= sq1;
		    DEFAULTNODE.stdio_quiet	= sq2;
		}
		else {
		    CONTEXT_SWITCH();
		    HANDLER(ne, timer, data);
		}
#else
		CONTEXT_SWITCH();
		HANDLER(ne, timer, data);
#endif
	    }
	    break;

	case SE_NODESTATE : {
	    np		= &(NP[THISNODE]);
	    switch ((int)np->runstate) {

	    case STATE_AUTOREBOOT :
		invoke_shutdown(THISNODE);	PING_TCLTK();
		invoke_reboot();
		break;

	    case STATE_UNDERREPAIR :
		if(dflag)
		    fprintf(stderr,"%s.repaired\n",np->nodename);

	    case STATE_CRASHED :
	    case STATE_REBOOTING :
		invoke_reboot();
		break;

	    case STATE_PAUSED :
		np->runstate	= STATE_RUNNING;
		if(dflag)
		    fprintf(stderr,"%s.resumed\n",np->nodename);
		break;

	    case STATE_RUNNING :
		np->runstate	= STATE_UNDERREPAIR;
		if(dflag)
		    fprintf(stderr,"%s.underrepair\n",np->nodename);

		filter_node_events(THISNODE);
		do_next = node_mttr();
		if(do_next != 0)
		    insert_event(SE_NODESTATE, EV_REBOOT, THISNODE,
				 TIMENOW_in_MSEC + do_next, NULLTIMER, 0);

		break;
	    }				/* end of switch ((int)np->runstate) */
	    DRAW_NODE_ICON();
	    break;
	}				/* end of case SE_NODESTATE */

	case SE_LINKSTATE : {
	    extern void report_linkstate(int);

	    int		l = THISNODE;	/* l is the link # that just changed */

	    if(!LP[l].linkup && ne == EV_REBOOT) {
		LP[l].linkup = TRUE;
		do_next	= link_mtbf(l);
		if(do_next != 0)
		    insert_event(SE_LINKSTATE, EV_SHUTDOWN, l,
				TIMENOW_in_MSEC + do_next, NULLTIMER, 0);
	    }
	    else if(LP[l].linkup && ne == EV_SHUTDOWN) {
		extern	int	clear_physical_link(int);
		extern	void	lose_all_drawframes(void);

		LP[l].linkup = FALSE;
		clear_physical_link(l);
#if	defined(USE_TCLTK)
		if(gattr.drawframes)
		    lose_all_drawframes();
#endif
		do_next	= link_mttr(l);
		if(do_next != 0)
		    insert_event(SE_LINKSTATE, EV_REBOOT, l,
				TIMENOW_in_MSEC + do_next, NULLTIMER, 0);
	    }
	    else
		break;
	    if(dflag)
		fprintf(stderr,"%s->%s.%s\n",
			NP[LP[l].endmin].nodename, NP[LP[l].endmax].nodename,
			LP[l].linkup ? "up" : "down" );
	    report_linkstate(l);
	    break;
	}				/* end of SE_LINKSTATE */

#if	defined(USE_TCLTK)
	case SE_DRAWFRAME : {
	    extern void		add_drawframe(DRAWFRAME *df);
	    extern void		free_drawframe(DRAWFRAME *df);
	    DRAWFRAME		*df;

	    np		= &(NP[THISNODE]);
	    df		= (DRAWFRAME *)data;
	    if(np->handler[(int)EV_DRAWFRAME]) {
		CONTEXT_SWITCH();
		HANDLER(EV_DRAWFRAME, NULLTIMER, (CnetData)&df->cdf);
		add_drawframe(df);
	    }
	    else
		free_drawframe(df);
	}

	case SE_PINGX : {
	    extern void		move_drawframes(void);

	    static long	next_inc	= PINGX_FREQ;
	    static long	last_ping	= 0L;
	    static long	next_stats	= 0L;
	    long	ms;

	    if(gattr.drawframes)
		move_drawframes();
	    PING_TCLTK();
	    CHECK_STATE_RETURN();
	    ms	= running_MSEC();

	    if(ms > next_stats) {
		int savevflag	= vflag;

		vflag		= FALSE;
		flush_stats();
		flush_nodestats();
		flush_linkstats();
		vflag		= savevflag;
		next_stats	= ms + STATS_FREQ;
	    }

	    if((ms - last_ping) < PINGX_FREQ)
		next_inc	+= (PINGX_FREQ/4);
	    else {
		next_inc	-= (PINGX_FREQ/4);
		if(next_inc <= 0)
		    next_inc	= 1;
	    }
	    last_ping	= ms;

	    insert_event(SE_PINGX, EV_NULL, ALLNODES,
			    TIMENOW_in_MSEC + next_inc, NULLTIMER, 0);
	    break;
	}
#endif

	case SE_REPORT : {
	    extern void	stats_summary(void);

	    if(dflag)
		fprintf(stderr,
		    "\tmsec=%5d contexts=%5d events=%6d nqueued=%3d\n",
				report_freq, cntx, events_handled, nqueued);
	    stats_summary();
	    DRAW_MAP();
	    events_handled	= 0;
	    cntx		= 0;
	    insert_event(SE_REPORT, EV_NULL, ALLNODES,
			    TIMENOW_in_MSEC + report_freq, NULLTIMER, 0);
	    break;
	}
	case SE_TIMESUP :
	    --TIMENOW_in_MSEC;
	    cnet_state	= STATE_TIMESUP;
	    return(events_left);
	}				/* end of switch( SE_? ) */
    }		/* while(GSQ.next) */
    return(events_left);
}


static void handle_linkstate(int l, int n)
{
    int	whichlink;

    np		= &(NP[n]);
    for(whichlink=0 ; whichlink<np->nlinks ; ++whichlink)
	if(l == np->links[whichlink])
	    break;
    if(n == SWAPPED_IN)
	linkinfo[whichlink+1].linkup	= LP[l].linkup;
    if(np->handler[ (int)EV_LINKSTATE ]) {
	THISNODE	= n;
	CONTEXT_SWITCH();
	HANDLER(EV_LINKSTATE, NULLTIMER, whichlink+1);
    }
}

void report_linkstate(int l)
{
    DRAW_LINK(l);

/*  ANNOUNCE LINKSTATE CHANGE TO EACH NODE, IF THEY WANT IT */

    handle_linkstate(l, LP[l].endmin);
    handle_linkstate(l, LP[l].endmax);
}


#if	defined(USE_TCLTK)
void single_event(int n, CnetEvent ev, CnetTimer timer, CnetData data)
{
    if(NP[n].runstate == STATE_RUNNING && NP[n].handler[(int)ev])  {
	THISNODE	= n;
	np		= &(NP[THISNODE]);
	CONTEXT_SWITCH();
	HANDLER(ev, timer, data);
    }
}
#endif	/* defined(USE_TCLTK) */
